/*
 * Omni.cpp
 *
 * Created: 2012/4/20 22:18:17
 *  Author: Andygongyb
 */ 

#include "Omni.hpp"

void COmniDir::set_dir(int16_t dir) {
	while (dir > 512) {
		dir -= 1024;
	}
	while (dir < -512) {
		dir += 1024;
	}
	int16_t tmp = dir + 128;
	if ((tmp >= 0) && (tmp < 512)) {
		// [-45, 135)
		tmp = 128 - dir;
		if (tmp < -512) {
			tmp += 1024;
		}
		m_north->set_servo(tmp);
		m_south->set_servo(tmp);
	} else {
		tmp = - dir - 384;
		if (tmp < -512) {
			tmp += 1024;
		}
		m_north->set_servo(tmp);
		tmp = - dir - 384;
		if (tmp < -512) {
			tmp += 1024;
		}
		m_south->set_servo(tmp);
	}
	tmp = dir + 384;
	if ((tmp >= 0) && (tmp < 512)) {
		// [-135, 45)
		tmp = - dir - 128;
		if (tmp < -512) {
			tmp += 1024;
		}
		m_east->set_servo(tmp);
		m_west->set_servo(tmp);
	} else {
		tmp = - dir + 384;
		if (tmp > 512) {
			tmp -= 1024;
		}
		m_east->set_servo(tmp);
		tmp = - dir + 384;
		if (tmp > 512) {
			tmp -= 1024;
		}
		m_west->set_servo(tmp);
	}
	m_dir = dir;
}

void COmniDir::move(int16_t speed) {
	int16_t tmp = m_dir + 128;
	if ((tmp >= 0) && (tmp < 512)) {
		// [-45, 135)
		m_north->set_wheel(speed);
		m_south->set_wheel(-speed);
	} else {
		m_north->set_wheel(-speed);
		m_south->set_wheel(speed);
	}
	tmp = m_dir + 384;
	if ((tmp >= 0) && (tmp < 512)) {
		// [-135, 45)
		m_west->set_wheel(speed);
		m_east->set_wheel(-speed);
	} else {
		m_west->set_wheel(-speed);
		m_east->set_wheel(speed);
	}
}

void COmniDir::move_pid(int8_t pid_set) {
	int16_t tmp = m_dir + 128;
	if ((tmp >= 0) && (tmp < 512)) {
		// [-45, 135)
		m_north->set_wheel_pid(pid_set);
		m_south->set_wheel_pid(-pid_set);
	} else {
		m_north->set_wheel_pid(-pid_set);
		m_south->set_wheel_pid(pid_set);
	}
	tmp = m_dir + 384;
	if ((tmp >= 0) && (tmp < 512)) {
		// [-135, 45)
		m_west->set_wheel_pid(pid_set);
		m_east->set_wheel_pid(-pid_set);
	} else {
		m_west->set_wheel_pid(-pid_set);
		m_east->set_wheel_pid(pid_set);
	}
}

void COmniDir::wheel_break(void) {
	m_north->wheel_break();
	m_south->wheel_break();
	m_east->wheel_break();
	m_west->wheel_break();
}

const float DT = 0.05;
const float KD = 6.0;
const float KP = 80.0;

void COmniDir::update(void) {
	m_north.m_worm_pos = m_north->get_adc();
	m_south.m_worm_pos = m_south->get_adc();
	m_east.m_worm_pos = m_east->get_adc();
	m_west.m_worm_pos = m_west->get_adc();
	m_mpu->get_euler(m_angle);
	bool need_adjust = false;
	float adjust = m_angle[0] * KP + ((m_angle[0] - m_pre_x) / DT) * KD;
	if (m_angle[0] > 2) {
		need_adjust = true;
		/*if (m_north.m_worm_pos > (m_north->m_pos_thr_l + 20)) {
			m_north->set_worm(adjust *2);
		} else {
			m_south->set_worm(-adjust);
			m_north->set_worm(adjust * 2);
		}*/
		m_south->set_worm(-adjust);
	} else if (m_angle[0] < -2) {
		need_adjust = true;
		/*if (m_south.m_worm_pos > (m_south->m_pos_thr_l + 20)) {
			m_south->set_worm(-adjust * 2);
		} else {
			m_north->set_worm(adjust);
			m_south->set_worm(-adjust * 2);
		}*/
		m_north->set_worm(adjust);
	}
	adjust = m_angle[1] * KP + ((m_angle[1] - m_pre_y) / DT) * KD;
	if (m_angle[1] > 2) {
		need_adjust = true;
		/*if (m_east.m_worm_pos > (m_east->m_pos_thr_l + 20)) {
			m_east->set_worm(adjust * 2);
		} else {
			m_west->set_worm(-adjust);
			m_east->set_worm(adjust * 2);
		}*/
		m_west->set_worm(-adjust);
	} else if (m_angle[1] < -2) {
		need_adjust = true;
		/*if (m_west.m_worm_pos > (m_west->m_pos_thr_l + 20)) {
			m_west->set_worm(-adjust * 2);
		} else {
			m_east->set_worm(adjust);
			m_west->set_worm(-adjust * 2);
		}*/
		m_east->set_worm(adjust);
	}
	if (need_adjust == false) {
		/*m_east->worm_break();
		m_west->worm_break();
		m_south->worm_break();
		m_north->worm_break();*/
		m_east->set_worm(1000);
		m_west->set_worm(1000);
		m_south->set_worm(1000);
		m_north->set_worm(1000);
	}
	m_pre_x = m_angle[0];
	m_pre_y = m_angle[1];
}