/*
 * SPI.hpp
 *
 *  Created on: Apr 8, 2012
 *      Author: Andygongyb
 */

#ifndef SPI_HPP_
#define SPI_HPP_

#include <cstddef>
#include <stdint.h>
#include <cstring>

extern "C" {
#include "gpio.h"
}

#include "PDCA.hpp"
#include "DataBlock.hpp"
#include "Initialization.hpp"

extern "C" void spi_pdca_irq(void) __attribute__((__interrupt__));

class CSPIDevice;

typedef class CSPI: public CPDCA_Peripheral {
public:
	typedef enum {
		CH0 = 0x0E, CH1 = 0x0D, CH2 = 0x0B, CH3 = 0x07,
	} channel_t;

	typedef struct {
		channel_t channel;
		uint8_t *w_buffer;
		uint16_t buffer_len;
		uint8_t *r_buffer;
		call_back_t before;
		call_back_t after;
		volatile flag_t *if_done;
	} packet_t;

private:
	typedef class CBlock: public CData {
	private:
		packet_t m_packet;

	public:
		CBlock(packet_t *packet = NULL);
		friend class CSPI;
		friend void spi_pdca_irq(void);
	}*pCBlock;

	CPDCA *m_opt_pdca;

	CBlock *m_head;
	CBlock *m_tail;
	uint16_t m_len;

	static volatile bool m_if_busy;

	static CSPI *m_instance;

	CSPI(void);

	inline void start(packet_t* packet) __attribute__((always_inline));

	inline void next_block(flag_t if_done) __attribute__((always_inline));

public:
	static CSPI *instance(void) {
		if (!m_instance)
			m_instance = new CSPI();
		return m_instance;
	}

	virtual ~CSPI(void);

	void append(packet_t* packet, bool if_need_free = false);

	void last_transfer(void) {
		AVR32_SPI0.CR.lastxfer = 1;
	}

	void test(bool if_on) {
		if (if_on) {
			AVR32_SPI0.MR.llb = 1;
		} else {
			AVR32_SPI0.MR.llb = 0;
		}
	}

	bool if_busy(void) {
		return m_if_busy;
	}

	uint32_t get_queue_len(void) {
		return m_len;
	}

	friend void spi_pdca_irq(void);
	friend class CSPIDevice;

}*pCSPI;

typedef class CSPIDevice {
public:

	typedef union {
		struct {
			uint8_t delay_between_transfers :8;
			uint8_t delay_before_spck :8;
			uint8_t baud_rate_neglect :8;
			uint8_t bits_per_transfer_neglect :4;
			bool chip_select_active_after_transfer :1;
			bool chip_select_not_active_after_transfer :1;
			bool clock_phase :1;
			bool clock_polarity :1;
		} st;
		uint32_t u32;
	} avr32_spi_csr_t;

	typedef struct {
		avr32_spi_csr_t csr;
		uint8_t data_size;
		uint32_t baud;
	} property_t;

protected:

	CSPI::channel_t const m_channel;
	static property_t m_property[4];
	static uint8_t m_npcs[4][2];

	volatile avr32_spi_csr_t *m_csr;
	property_t *m_the_property;

	CSPIDevice(CSPI::channel_t channel, property_t *property = NULL);

	virtual ~CSPIDevice(void);

	void set_property(property_t &property);
	void get_property(property_t &property) {
		std::memcpy(&property, m_the_property, sizeof(property_t));
	}

	void append(CSPI::packet_t* packet, bool if_need_free = false) {
		CSPI::instance()->append(packet, if_need_free);
	}

public:

}*pCSPIDevice;

#endif /* SPI_HPP_ */
