/*
 * EIC.cpp
 *
 *  Created on: Nov 8, 2011
 *      Author: Andygongyb
 */

#include "EIC.hpp"

const char *CExtInt_exception::m_channel_in_use =
		"External interrupt channel in use";

CExtInt *CExtInt::m_instance = NULL;

CExtInt::property_t CExtInt::m_property[8] = { { CExtInt::RISING_EDGE, false,
		false }, { CExtInt::RISING_EDGE, false, false }, {
		CExtInt::RISING_EDGE, false, false }, { CExtInt::RISING_EDGE, false,
		false }, { CExtInt::RISING_EDGE, false, false }, {
		CExtInt::RISING_EDGE, false, false }, { CExtInt::RISING_EDGE, false,
		false }, { CExtInt::RISING_EDGE, false, false } };

void (*( CExtInt::m_handler[8]))(uint8_t) = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};

__attribute__((__section__(".eicmem"))) const uint16_t int_map[8][3] = { { AVR32_EIC_EXTINT_1_1_PIN,
		AVR32_EIC_EXTINT_1_1_FUNCTION, AVR32_EIC_IRQ_1 }, {
		AVR32_EIC_EXTINT_2_2_PIN, AVR32_EIC_EXTINT_2_2_FUNCTION,
		AVR32_EIC_IRQ_2 }, { AVR32_EIC_EXTINT_3_1_PIN,
		AVR32_EIC_EXTINT_3_1_FUNCTION, AVR32_EIC_IRQ_3 }, {
		AVR32_EIC_EXTINT_4_1_PIN, AVR32_EIC_EXTINT_4_1_FUNCTION,
		AVR32_EIC_IRQ_4 }, { AVR32_EIC_EXTINT_5_PIN,
		AVR32_EIC_EXTINT_5_FUNCTION, AVR32_EIC_IRQ_5 }, {
		AVR32_EIC_EXTINT_6_1_PIN, AVR32_EIC_EXTINT_6_1_FUNCTION,
		AVR32_EIC_IRQ_6 }, { AVR32_EIC_EXTINT_7_1_PIN,
		AVR32_EIC_EXTINT_7_1_FUNCTION, AVR32_EIC_IRQ_7 }, {
		AVR32_EIC_EXTINT_8_1_PIN, AVR32_EIC_EXTINT_8_1_FUNCTION,
		AVR32_EIC_IRQ_8 } };

typedef void (*handler_t)(void);
handler_t handler_map[8] = {
		&ext_int_handler_1,
		&ext_int_handler_2,
		&ext_int_handler_3,
		&ext_int_handler_4,
		&ext_int_handler_5,
		&ext_int_handler_6,
		&ext_int_handler_7,
		&ext_int_handler_8
};

CExtInt::CExtInt(void) {

	bool int_sta = Is_global_interrupt_enabled();
	Disable_global_interrupt();

	for (uint8_t i = 0; i < 8; i++) {
		set_property(static_cast<int_num_t> (i), m_property[i]);
		INTC_register_interrupt(handler_map[i], int_map[i][2], AVR32_INTC_INT1);
	}
	if (int_sta)
		Enable_global_interrupt();

}

CExtInt::~CExtInt(void) {
	bool int_sta = Is_global_interrupt_enabled();
	Disable_global_interrupt();

	for (uint8_t i = 0; i < 8; i++) {
		detach(static_cast<int_num_t> (i));
	}

	m_instance = NULL;

	if (int_sta)
		Enable_global_interrupt();
}

void CExtInt::attach(int_num_t int_num, void(*handler)(uint8_t),
		property_t *property) {
	if (m_handler[int_num] != NULL) {
		throw CExtInt_exception(CExtInt_exception::CHANNEL_IN_USE);
	}

	bool int_sta = Is_global_interrupt_enabled();
	Disable_global_interrupt();

	if (property) {
		set_property(int_num, *property);
	}

	m_handler[int_num] = handler;
	gpio_enable_module_pin(int_map[int_num][0], int_map[int_num][1]); // setup I/O function

	AVR32_EIC.en = 1 << (int_num + 1); // enable the EIC channel

	if (int_sta)
		Enable_global_interrupt();
}

void CExtInt::detach(int_num_t int_num) {
	AVR32_EIC.dis = 1 << (int_num + 1); // disable the EIC channel
	m_handler[int_num] = NULL;

	gpio_enable_gpio_pin( int_map[int_num][0]); // setup I/O back to gpio
}

void CExtInt::set_property(int_num_t int_num, property_t &property) {
	bool int_sta = Is_global_interrupt_enabled();
	Disable_global_interrupt();

	m_property[int_num].mode = property.mode;
	m_property[int_num].async = property.async;
	m_property[int_num].filter = property.filter;
	uint8_t real_int_num = int_num + 1;
	if (property.mode & 0x02) {
		AVR32_EIC.mode |= 1 << real_int_num;
		if (property.mode & 0x01) {
			AVR32_EIC.level |= 1 << real_int_num;
		} else {
			AVR32_EIC.level &= ~(1 << real_int_num);
		}
	} else {
		AVR32_EIC.mode &= ~(1 << real_int_num);
		if (property.mode & 0x01) {
			AVR32_EIC.edge |= 1 << real_int_num;
		} else {
			AVR32_EIC.edge &= ~(1 << real_int_num);
		}
	}
	if (property.async) {
		AVR32_EIC.async |= 1 << real_int_num;
	} else {
		AVR32_EIC.async &= ~(1 << real_int_num);
	}
	if (property.filter) {
		AVR32_EIC.filter |= 1 << real_int_num;
	} else {
		AVR32_EIC.filter &= ~(1 << real_int_num);
	}
	if (int_sta)
		Enable_global_interrupt();
}

void ext_int_handler_1(void) {
	AVR32_EIC.icr = 1 << 1; // clear the flag
	(*CExtInt::m_handler[0])(1);
}

void ext_int_handler_2(void) {
	AVR32_EIC.icr = 1 << 2; // clear the flag
	(*CExtInt::m_handler[1])(2);
}

void ext_int_handler_3(void) {
	AVR32_EIC.icr = 1 << 3; // clear the flag
	(*CExtInt::m_handler[2])(3);
}

void ext_int_handler_4(void) {
	AVR32_EIC.icr = 1 << 4; // clear the flag
	(*CExtInt::m_handler[3])(4);
}

void ext_int_handler_5(void) {
	AVR32_EIC.icr = 1 << 5; // clear the flag
	(*CExtInt::m_handler[4])(5);
}

void ext_int_handler_6(void) {
	AVR32_EIC.icr = 1 << 6; // clear the flag
	(*CExtInt::m_handler[5])(6);
}

void ext_int_handler_7(void) {
	AVR32_EIC.icr = 1 << 7; // clear the flag
	(*CExtInt::m_handler[6])(7);
}

void ext_int_handler_8(void) {
	AVR32_EIC.icr = 1 << 8; // clear the flag
	(*CExtInt::m_handler[7])(8);
}