/*
 * EIC.hpp
 *
 *  Created on: Nov 8, 2011
 *      Author: Andygongyb
 */

#ifndef EIC_HPP_
#define EIC_HPP_

extern "C" {
#include "gpio.h"
#include "intc.h"
}

#include <stdint.h>
#include <cstddef>

#include "exception.hpp"

extern "C" void ext_int_handler_1(void)__attribute__((__interrupt__));
extern "C" void ext_int_handler_2(void)__attribute__((__interrupt__));
extern "C" void ext_int_handler_3(void)__attribute__((__interrupt__));
extern "C" void ext_int_handler_4(void)__attribute__((__interrupt__));
extern "C" void ext_int_handler_5(void)__attribute__((__interrupt__));
extern "C" void ext_int_handler_6(void)__attribute__((__interrupt__));
extern "C" void ext_int_handler_7(void)__attribute__((__interrupt__));
extern "C" void ext_int_handler_8(void)__attribute__((__interrupt__));

typedef class CExtInt_exception: public CException {
	static const char *m_channel_in_use;
public:

	typedef enum {
		CHANNEL_IN_USE = -1
	} errno_t;

	CExtInt_exception(errno_t errno) :
		CException(NULL, 0) {
		const char *name = NULL;
		switch (errno) {
		case CHANNEL_IN_USE:
			name = m_channel_in_use;
			break;
		}
		strcpy(m_name, name);
		m_errno = errno;
	}
}*pCExtInt_exception;

typedef class CExtInt {
public:

	typedef enum {
		FALLING_EDGE = 0, RISING_EDGE = 1, LOW_LEVEL = 2, HIGH_LEVEL = 3,
	} mode_t;

	typedef enum {
		EXTINT1 = 0,
		EXTINT2 = 1,
		EXTINT3 = 2,
		EXTINT4 = 3,
		EXTINT5 = 4,
		EXTINT6 = 5,
		EXTINT7 = 6,
		EXTINT8 = 7,
	} int_num_t;

	typedef struct {
		mode_t mode :2;
		bool filter :1;
		bool async :1;
	} property_t;

private:
	static property_t m_property[8];
	static void (*( m_handler[8]))(uint8_t);

	static CExtInt *m_instance;

	CExtInt(void);

public:

	static CExtInt *instance(void) {
		if (!m_instance)
			m_instance = new CExtInt();
		return m_instance;
	}

	~CExtInt(void);

	void attach(int_num_t int_num, void(*handler)(uint8_t),
			property_t *property = NULL);
	void detach(int_num_t int_num);

	void set_property(int_num_t int_num, property_t &property);

	void en_int(int_num_t int_num) {
		AVR32_EIC.ier = 1 << (int_num + 1);
	}

	void dis_int(int_num_t int_num) {
		AVR32_EIC.idr = 1 << (int_num + 1);
	}

	friend void ext_int_handler_1(void);
	friend void ext_int_handler_2(void);
	friend void ext_int_handler_3(void);
	friend void ext_int_handler_4(void);
	friend void ext_int_handler_5(void);
	friend void ext_int_handler_6(void);
	friend void ext_int_handler_7(void);
	friend void ext_int_handler_8(void);
}*pCExtInt;

#endif /* EIC_HPP_ */
