/*
 * TWI_ASYN.cpp
 *
 *  Created on: Apr 9, 2012
 *      Author: Andygongyb
 */

#include "TWI_ASYN.hpp"

template<CTWIM::module_t N>
CTWIM_ch<N>* CTWIM_ch<N>::m_instance = NULL;

CTWIM::CCMD::CCMD(packet_t* packet) {
	if (packet) {
		memcpy(&m_packet, packet, sizeof(packet_t));
		if (packet->if_done)
			*(packet->if_done) = PENDING;
	} else {
		memset(&m_packet, 0, sizeof(packet_t));
	}
}

CTWIM::CTWIM(module_t module) {
	bool int_sta = Is_global_interrupt_enabled();
	Disable_global_interrupt();
	
	m_mutex = xSemaphoreCreateMutex();

	m_head = new CCMD();
	m_tail = m_head;
	m_len = 0;

	m_pdca = NULL;
	m_if_busy = false;

	CInit::instance();

	if (module == TWIM0) {
		m_twim = &AVR32_TWIM0;
		gpio_enable_module_pin(AVR32_TWIMS0_TWCK_PIN,
				AVR32_TWIMS0_TWCK_FUNCTION);
		gpio_enable_module_pin(AVR32_TWIMS0_TWD_PIN, AVR32_TWIMS0_TWD_FUNCTION);
		m_tx_module = CPDCA::TWIM0_TX;
		m_handler = &twim0_pdca_irq;
		INTC_register_interrupt(&twim0_irq, AVR32_TWIM0_IRQ, AVR32_INTC_INT3);
	} else {
		m_twim = &AVR32_TWIM1;
		gpio_enable_module_pin(AVR32_TWIMS1_TWCK_PIN,
				AVR32_TWIMS1_TWCK_FUNCTION);
		gpio_enable_module_pin(AVR32_TWIMS1_TWD_PIN, AVR32_TWIMS1_TWD_FUNCTION);
		m_tx_module = CPDCA::TWIM1_TX;
		m_handler = &twim1_pdca_irq;
		INTC_register_interrupt(&twim1_irq, AVR32_TWIM1_IRQ, AVR32_INTC_INT3);
	}

	m_twim->cr = AVR32_TWIM_CR_MEN_MASK | AVR32_TWIM_CR_SMDIS_MASK; // disable SMBus mode and enable master mode
	// Reset TWI
	m_twim->cr = AVR32_TWIM_CR_SWRST_MASK;
	// Clear all status
	m_twim->scr = ~0UL;

	// Select the speed
	set_freq(400000UL);

	// mask all interrupts
	m_twim->idr = ~0UL;
	m_twim->ier = AVR32_TWIM_IER_ANAK_MASK | AVR32_TWIM_IER_DNAK_MASK
			| AVR32_TWIM_IER_ARBLST_MASK | AVR32_TWIM_IER_CCOMP_MASK;

	if (int_sta)
		Enable_global_interrupt();
}

void CTWIM::set_freq(uint32_t clk) {
	uint8_t cwgr_exp = 0;
	uint32_t f_prescaled = CInit::instance()->get_pba_freq() / clk / 2;
	// f_prescaled must fit in 8 bits, cwgr_exp must fit in 3 bits
	while ((f_prescaled > 0xFF) && (cwgr_exp <= 0x7)) {
		// increase clock divider
		cwgr_exp++;
		// divide f_prescaled value
		f_prescaled /= 2;
	}
	// set clock waveform generator register
	m_twim->cwgr = ((f_prescaled / 2) << AVR32_TWIM_CWGR_LOW_OFFSET)
			| ((f_prescaled - f_prescaled / 2) << AVR32_TWIM_CWGR_HIGH_OFFSET)
			| (cwgr_exp << AVR32_TWIM_CWGR_EXP_OFFSET) | (3
			<< AVR32_TWIM_CWGR_DATA_OFFSET) | (f_prescaled
			<< AVR32_TWIM_CWGR_STASTO_OFFSET);
}

inline void CTWIM::start(packet_t* packet) {

	// Reset the TWIM module to clear the THR register
	m_twim->cr = AVR32_TWIM_CR_MEN_MASK;
	m_twim->cr = AVR32_TWIM_CR_SWRST_MASK;
	m_twim->cr = AVR32_TWIM_CR_MDIS_MASK;
	// clear all status
	m_twim->scr = ~0UL;
	CPDCA::property_t pdca_property = { CPDCA::BYTE, m_tx_module };
	m_pdca->set_property(pdca_property);
	if (packet->mode) {
		// read mode
		m_read_pack = packet;
		m_pdca->set_next_mem_loc(packet->buffer, 1);
		gpio_enable_gpio_pin(AVR32_PIN_PD12);
		gpio_set_gpio_pin(AVR32_PIN_PD12);	

		m_twim->cmdr = (packet->slave_addr << AVR32_TWIM_CMDR_SADR_OFFSET)
				| AVR32_TWIM_CMDR_START_MASK | (1
				<< AVR32_TWIM_CMDR_NBYTES_OFFSET) | AVR32_TWIM_CMDR_VALID_MASK;
		m_twim->ncmdr = 0x01 | (packet->slave_addr
				<< AVR32_TWIM_CMDR_SADR_OFFSET) | AVR32_TWIM_CMDR_START_MASK
				| AVR32_TWIM_CMDR_STOP_MASK | ((packet->buffer_len - 1)
				<< AVR32_TWIM_CMDR_NBYTES_OFFSET) | AVR32_TWIM_CMDR_VALID_MASK;
	} else {
		m_read_pack = NULL;
		m_pdca->set_next_mem_loc(packet->buffer, packet->buffer_len);

		m_twim->cmdr = (packet->slave_addr << AVR32_TWIM_CMDR_SADR_OFFSET)
				| AVR32_TWIM_CMDR_START_MASK | AVR32_TWIM_CMDR_STOP_MASK
				| (packet->buffer_len << AVR32_TWIM_CMDR_NBYTES_OFFSET)
				| AVR32_TWIM_CMDR_VALID_MASK;
	}

	m_pdca->on();

	m_twim->cr = AVR32_TWIM_CR_MEN_MASK;
}

inline portBASE_TYPE CTWIM::next_cmd(volatile flag_t if_done) {
	CCMD *tmp = m_head;
	m_head = (CCMD*) (m_head->m_next);
	portBASE_TYPE xYieldRequired = pdFALSE;
	
	uint8_t *last_data = m_head->m_packet.buffer;
	uint16_t last_len = m_head->m_packet.buffer_len;

	if (m_head->m_packet.after) {
		(*(m_head->m_packet.after))(last_data, last_len);
	}

	if (m_head->m_packet.if_done)
		*(m_head->m_packet.if_done) = if_done;
	if (m_head->m_packet.task) {
		xYieldRequired = xTaskResumeFromISR(m_head->m_packet.task);
	}
	CDataControl::instance()->wait_for_free(tmp);
	m_len--;
	if (m_len) {
		m_pdca->off();
		if (((CCMD*) (m_head->m_next))->m_packet.before) {
			(*((CCMD*) (m_head->m_next))->m_packet.before)(last_data, last_len);
		}
		start(&((CCMD*) (m_head->m_next))->m_packet);
	} else {
		// no more commend, finish the transmission
		m_pdca->dis_all_int();
		detach_pdca( m_pdca);
		m_if_busy = false;
	}
	return xYieldRequired;
}

void CTWIM::append(packet_t* packet, bool need_free, bool need_block) {
	xSemaphoreTake(m_mutex, portMAX_DELAY);
	
	volatile flag_t block;
	if ((xTaskGetSchedulerState() == taskSCHEDULER_RUNNING) && need_block) {
		packet->task = xTaskGetCurrentTaskHandle();
	} else {
		packet->task = NULL;
		if (packet->if_done == NULL) {
			packet->if_done = &block;
		}
	}
	
	m_tail->m_next = new CCMD(packet);
	m_tail = (CCMD*) (m_tail->m_next);

	if (need_free)
		m_tail->m_sub_buffer = packet->buffer;
	m_len++;
	
	portENTER_CRITICAL();
	if (!m_if_busy) {
		m_if_busy = true;

		attach_pdca( m_pdca);
		m_pdca->set_handler(m_handler);
		m_pdca->dis_all_int();
		m_pdca->en_err_int();
		m_pdca->get_channel()->isr;

		if (m_tail->m_packet.before) {
			(*(m_tail->m_packet.before))(NULL, 0);
		}
		start(&m_tail->m_packet);
	}
	if (packet->task) {
		// suspend the calling task in a critical session in case the twim interrupt
		// may come first
		// when switching context, critical nesting will be written for each task
		vTaskSuspend(NULL);
	}
	portEXIT_CRITICAL();
	xSemaphoreGive(m_mutex);
	if (need_block && (packet->task == NULL)) {
		do {} while (*packet->if_done == PENDING);
	}
}

void twim0_pdca_irq(void) {
	portENTER_SWITCHING_ISR();
	portLONG need_switch = pdFALSE;
	CTWIM* the_twim = CTWIM_ch<CTWIM::TWIM0>::instance();
	volatile avr32_pdca_channel_t *the_pdca;
	the_pdca = the_twim->m_pdca->get_channel();
	uint32_t status = the_pdca->isr;

	if (status & AVR32_PDCA_ISR0_TERR_MASK) {
		the_twim->m_pdca->clear_err();
		need_switch = the_twim->next_cmd(FAILED);
	}
	asm volatile (
		"mov r12, %0"
		:
		: "r"(need_switch)
		: "r12"
	);
	portEXIT_SWITCHING_ISR();
}

void twim1_pdca_irq(void) {
	portENTER_SWITCHING_ISR();
	portLONG need_switch = pdFALSE;
	CTWIM* the_twim = CTWIM_ch<CTWIM::TWIM1>::instance();
	volatile avr32_pdca_channel_t *the_pdca;
	the_pdca = the_twim->m_pdca->get_channel();
	uint32_t status = the_pdca->isr;

	if (status & AVR32_PDCA_ISR0_TERR_MASK) {
		the_twim->m_pdca->clear_err();
		need_switch = the_twim->next_cmd(FAILED);
	}
	asm volatile (
		"mov r12, %0"
		:
		: "r"(need_switch)
		: "r12"
	);
	portEXIT_SWITCHING_ISR();
}

void twim0_irq(void) {
	portENTER_SWITCHING_ISR();
	portLONG need_switch = pdFALSE;
	uint32_t status = AVR32_TWIM0.sr;
	AVR32_TWIM0.scr = ~0UL;
	CTWIM* the_twim = CTWIM_ch<CTWIM::TWIM0>::instance();
	if (status & AVR32_TWIM_SR_CCOMP_MASK) {
		if (the_twim->m_read_pack) {	
			CPDCA::property_t pdca_property = { CPDCA::BYTE, CPDCA::TWIM0_RX };
			the_twim->m_pdca->set_property(pdca_property);
			the_twim->m_pdca->set_next_mem_loc(the_twim->m_read_pack->buffer + 1, 
					the_twim->m_read_pack->buffer_len - 1);
			the_twim->m_read_pack = NULL;
		} else {
			need_switch = the_twim->next_cmd(DONE);
		}	
	}
	asm volatile (
		"mov r12, %0"
		:
		: "r"(need_switch)
		: "r12"
	);
	portEXIT_SWITCHING_ISR();
}

void twim1_irq(void) {
	portENTER_SWITCHING_ISR();
	portLONG need_switch = pdFALSE;
	uint32_t status = AVR32_TWIM1.sr;
	AVR32_TWIM1.scr = ~0UL;
	CTWIM* the_twim = CTWIM_ch<CTWIM::TWIM1>::instance();
	if (status & AVR32_TWIM_SR_CCOMP_MASK) {
		if (the_twim->m_read_pack) {
			CPDCA::property_t pdca_property = { CPDCA::BYTE, CPDCA::TWIM1_RX };
			the_twim->m_pdca->set_property(pdca_property);
			the_twim->m_pdca->set_next_mem_loc(the_twim->m_read_pack->buffer + 1, 
					the_twim->m_read_pack->buffer_len - 1);
			the_twim->m_read_pack = NULL;
		} else {
			need_switch = the_twim->next_cmd(DONE);
		}	
	}
	asm volatile (
		"mov r12, %0"
		:
		: "r"(need_switch)
		: "r12"
	);
	portEXIT_SWITCHING_ISR();
}

void CTWIM_device::reg_w(uint8_t reg, uint8_t val) {
	uint8_t buffer[2];
	buffer[0] = reg;
	buffer[1] = val;
	CTWIM::packet_t packet = {CTWIM::WRITE, m_addr, buffer, 2};
	m_twim->append(&packet, false, true);
}

void CTWIM_device::write_seq (uint8_t* buffer, uint8_t len) {
	CTWIM::packet_t packet = {CTWIM::WRITE, m_addr, buffer, len};
	m_twim->append(&packet, false, true);
}

uint8_t CTWIM_device::reg_r (uint8_t reg) {
	uint8_t buffer[2];
	buffer[0] = reg;
	CTWIM::packet_t packet = {CTWIM::READ, m_addr, buffer, 2};
	m_twim->append(&packet, false, true);
	return buffer[1];
}

void CTWIM_device::read_seq (uint8_t* buffer, uint8_t len) {
	CTWIM::packet_t packet = {CTWIM::READ, m_addr, buffer, len};
	m_twim->append(&packet, false, true);
}