/*
 * ECE445_slave.c
 *
 * Created: 2012/4/16 14:28:09
 *  Author: Andygongyb
 */ 


#include <avr/io.h>
#include <util/delay.h>
#include <stdint.h>
#include <util/twi.h>
#include <avr/interrupt.h>

#include "wheel.h"
#include "timer1.h"
#include "worm.h"
#include "servo.h"
#include "twi.h"
#include "adc.h"

#define LED0_ON PORTB &= ~_BV(PB7)
#define LED0_OFF PORTB |= _BV(PB7)
#define LED1_ON PORTD &= ~_BV(PD4)
#define LED1_OFF PORTD |= _BV(PD4)

typedef enum {
	NO_CMD = 0x00U,
	SET_WHEEL = 0x01U,
	SET_WORM = 0x02U,
	SET_SERVO = 0x03U,
	SET_SERVO_PRO = 0x04U,
	SET_ADC_THR = 0x05U,
	WORM_BREAK = 0x06U,
	WHEEL_BREAK = 0x07U,
	SET_WHEEL_PID = 0x08U,
	GET_WHEEL_SPEED = 0xFEU,
	GET_ADC = 0xFFU
} twi_cmd_t;

#define READ_THR 128U

#define SET_WHEEL_BIT _BV(SET_WHEEL)
#define SET_WORM_BIT _BV(SET_WORM)
#define SET_SERVO_BIT _BV(SET_SERVO)
#define SET_SERVO_PRO_BIT _BV(SET_SERVO_PRO)
#define SET_ADC_THR_BIT _BV(SET_ADC_THR)
#define WORM_BREAK_BIT _BV(WORM_BREAK)
#define WHEEL_BREAK_BIT _BV(WHEEL_BREAK)
#define SET_WHEEL_PID_BIT _BV(SET_WHEEL_PID)

static twi_cmd_t twi_cmd = NO_CMD;
volatile uint16_t twi_cmd_map = 0;

static uint8_t cmd_wheel_cnt;
static uint8_t cmd_worm_cnt;
static uint8_t cmd_servo_cnt;
static uint8_t cmd_servo_pro_cnt;
static uint8_t cmd_adc_thr_cnt;

static uint8_t adc_h;
static uint8_t wheel_spd_h;
static int8_t pid_buffer;

typedef union {
	uint8_t received[2];
	int16_t actual;
} cmd_buffer_t;

cmd_buffer_t wheel_buffer, worm_buffer, servo_buffer;
uint8_t adc_buffer[4];

int main(void)
{
	PORTD |= _BV(PD4);
	DDRD |= _BV(PD4);
	PORTB |= _BV(PB7);
	DDRB |= _BV(PB7);

	twi_init();
	adc_init();
	timer1_init();
	wheel_init();
	worm_init();
	servo_init();
	timer1_start();
	servo_set(0);
	worm_break();
	wheel_break();
	wheel_buffer.actual = 0;
	servo_buffer.actual = 0;
	worm_buffer.actual = 0;
    while(1)
    {
		while(!twi_cmd_map);
		if (twi_cmd_map & SET_WHEEL_PID_BIT) {
			wheel_pid_set(pid_buffer);
			twi_cmd_map &= ~SET_WHEEL_PID_BIT;
		}
		if (twi_cmd_map & SET_WORM_BIT) {
			worm_set(worm_buffer.actual);
			twi_cmd_map &= ~SET_WORM_BIT;
		}
		if (twi_cmd_map & SET_SERVO_BIT) {
			servo_set(servo_buffer.actual);
			twi_cmd_map &= ~SET_SERVO_BIT;
		}
		if (twi_cmd_map & SET_SERVO_PRO_BIT) {
			servo_set(servo_buffer.actual);
			twi_cmd_map &= ~SET_SERVO_PRO_BIT;
		}
		if (twi_cmd_map & SET_WHEEL_BIT) {
			wheel_set(wheel_buffer.actual);
			twi_cmd_map &= ~SET_WHEEL_BIT;
		}
		if (twi_cmd_map & SET_ADC_THR_BIT) {
			cli();
			memcpy(adc_property.buffer, adc_buffer, 4);
			sei();
			twi_cmd_map &= ~SET_ADC_THR_BIT;
		}
		if (twi_cmd_map & WORM_BREAK_BIT) {
			worm_break();
			twi_cmd_map &= ~WORM_BREAK_BIT;
		}
		if (twi_cmd_map & WHEEL_BREAK_BIT) {
			wheel_break();
			twi_cmd_map &= ~WHEEL_BREAK_BIT;
		}
    }
}

ISR(TWI_vect) {
	uint16_t tmp;
	switch (TW_STATUS) {
	case TW_SR_DATA_ACK:
		switch (twi_cmd) {
		case NO_CMD:
			twi_cmd = TWDR;
			LED0_ON;
			cmd_servo_cnt = 1;
			cmd_wheel_cnt = 1;
			cmd_worm_cnt = 1;
			cmd_servo_pro_cnt = 0;
			cmd_adc_thr_cnt = 3;
			break;
		case SET_SERVO:
			servo_buffer.received[cmd_servo_cnt] = TWDR;
			cmd_servo_cnt--;
			break;
		case SET_WHEEL:
			wheel_buffer.received[cmd_wheel_cnt] = TWDR;
			cmd_wheel_cnt--;
			break;
		case SET_WORM:
			worm_buffer.received[cmd_worm_cnt] = TWDR;
			cmd_worm_cnt--;
			break;
		case SET_SERVO_PRO:
			servo_property.buffer[cmd_servo_pro_cnt] = TWDR;
			cmd_servo_pro_cnt++;
			break;
		case SET_ADC_THR:
			adc_buffer[cmd_adc_thr_cnt] = TWDR;
			cmd_adc_thr_cnt--;
			break;
		case SET_WHEEL_PID:
			pid_buffer = TWDR;
			break;
		default:
			break;
		}
		TWCR = _BV(TWIE) | _BV(TWINT) | _BV(TWEA) | _BV(TWEN);
		break;
	case TW_ST_SLA_ACK:
		LED1_ON;
		switch (twi_cmd) {
		case GET_ADC:
			tmp = adc_get();
			adc_h = tmp & 0xFF;
			TWDR = (tmp >> 8) & 0xFF;
			break;
		case GET_WHEEL_SPEED:
			tmp = wheel_get_counter();
			wheel_spd_h = tmp & 0xFF;
			TWDR = (tmp >> 8) & 0xFF;
		default:
			break;
		}
		TWCR = _BV(TWIE) | _BV(TWINT) | _BV(TWEA) | _BV(TWEN);
		break;
	case TW_ST_DATA_ACK:
		switch (twi_cmd) {
		case GET_ADC:
			TWDR = adc_h;
			break;
		case GET_WHEEL_SPEED:
			TWDR = wheel_spd_h;
			break;
		default:
			break;
		}
		TWCR = _BV(TWIE) | _BV(TWINT) | _BV(TWEN);
		break;
	case TW_ST_DATA_NACK:
	case TW_ST_LAST_DATA:
		twi_cmd = NO_CMD;
		TWCR = _BV(TWIE) | _BV(TWINT) | _BV(TWEA) | _BV(TWEN);
		LED1_OFF;
		break;
	case TW_SR_STOP:
		if (twi_cmd <= READ_THR) {
			twi_cmd_map |= 1 << twi_cmd;
			twi_cmd = NO_CMD;
		}	
		LED0_OFF;
		TWCR = _BV(TWIE) | _BV(TWINT) | _BV(TWEA) | _BV(TWEN);
		break;
	case TW_BUS_ERROR:
		TWCR = _BV(TWIE) | _BV(TWINT) | _BV(TWEA) | _BV(TWEN) | _BV(TWSTO);
		break;
	default:
		TWCR = _BV(TWIE) | _BV(TWINT) | _BV(TWEA) | _BV(TWEN);
	}
}